/*

This file is part of jHaushalt.

jHaushalt is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

jHaushalt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with jHaushalt; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

(C)opyright 2002-2004 Dr. Lars H. Hahn

*/

package haushalt.daten;

import haushalt.daten.zeitraum.AbstractZeitraum;

import java.io.DataInputStream;
import java.io.InputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.StringTokenizer;

/**
 * Die Datenbasis beinhaltet alle Buchungen und dient zum 
 * Zugriff auf alle Daten. Sie verwaltet die Register und 
 * Kategorien. In der Datenbasis sind auch die gemerkten
 * Buchungen gespeichert. 
 * @author Dr. Lars H. Hahn
 * @version 2.01/2004.08.25
 */

 /*
  * 2004.08.25 BugFix: Einfgen einer StandardBuchung liefert Position zurck.
  */
public class Datenbasis {
	private final boolean DEBUG = false;

	private final String version = "2.01";
	private boolean geaendert = false;

  public Datenbasis() {
		kategorieListe = new ArrayList();
		kategorieListe.add(Kategorie.SONSTIGES);
  }
  
  /**
   * Ist <code>true</code>, wenn Daten gendert wurden.
   * @return gendert oder nicht gendert
   */
  public boolean istGeaendert() {
  	return geaendert;
  }
  
	// -- Kategorien --------------------------------------------
	 
  private final ArrayList kategorieListe;

	/**
	 * Liefert die Kategorie mit dem angegebene Namen zurck.
	 * Wenn sie noch nicht existiert, wird sie erzeugt.
	 * @param name Name der Kategorie
	 * @param hauptkategorie Hauptkategorie
	 * @return gesuchte Kategorie
	 */
	public Kategorie findeOderErzeugeKategorie(String name, Kategorie hauptkategorie) {
		Kategorie kategorie = new Kategorie(name, hauptkategorie);
		int pos = Collections.binarySearch(kategorieListe, kategorie);
		if (pos < 0) {
			kategorieListe.add(-pos-1, kategorie);
			return kategorie;
		}
		return (Kategorie)kategorieListe.get(pos);
	}
	
	public Kategorie findeOderErzeugeKategorie(String vollerName) {
	  int n = vollerName.indexOf(":");
	  if(n == -1)
	    return findeOderErzeugeKategorie(vollerName, null);
	  Kategorie hauptkategorie = findeOderErzeugeKategorie(vollerName.substring(0,n), null);
	  return findeOderErzeugeKategorie(vollerName.substring(n+1), hauptkategorie);
	}

	public boolean isKategorie(String name, Kategorie hauptkategorie) {
		Kategorie kategorie = new Kategorie(name, hauptkategorie);
		int pos = Collections.binarySearch(kategorieListe, kategorie);
		if (pos < 0)
		  return false;
		return true;
	}

	/**
	 * Liefert ein Array mit allen Kategorien.
	 * Wenn die Unterkategorien verwendet werden sollen,
	 * werden alle Kategorien geliefert, sonst nur die
	 * Haupt-Kategorien
	 * @return Array mit den Kategorien
	 */
	public Kategorie[] getKategorien() {
		if(unterkategorienVerwenden)
      return (Kategorie[])kategorieListe.toArray(new Kategorie[kategorieListe.size()]);
		// Es sollen nur die Hauptkategorien ausgegeben werden:
		final int anzahl = kategorieListe.size();
    Kategorie[] kategorien = new Kategorie[anzahl];
		int katZaehler = 0;
		for(int i=0; i<anzahl; i++) {
			Kategorie kategorie = (Kategorie)kategorieListe.get(i);
			if(kategorie.isHauptkategorie())
				kategorien[katZaehler++] = kategorie;
		}
    Kategorie[] haupt = new Kategorie[katZaehler];
		System.arraycopy(kategorien,0,haupt,0,katZaehler);
		return haupt;
	}

	/**
   * Ersetzt in allen Registern die angegebne Kategorie durch eine neue.
	 * @param alteKategorie
	 * @param neueKategorie
	 * @return Anzahl ersetzter Kategorien
	 */
  public int ersetzeKategorie(Kategorie alteKategorie, Kategorie neueKategorie) {
    int zaehler = 0;
		for(int i=0; i<registerListe.size(); i++) {
			Register register = (Register)registerListe.get(i);
			for(int j=0;j<register.getAnzahlBuchungen();j++)
				zaehler += register.getBuchung(j).ersetzeKategorie(alteKategorie, neueKategorie);
		}
		geaendert = true;
		cacheAktuell = false;
		return zaehler;
	}

  // -- Register ----------------------------------------------

	private final ArrayList registerListe = new ArrayList();
	
  /**
   * Erzeugt ein Register mit dem angegebenen Namen.
   * Wenn es schon ein Register mit gleichem Namen gibt, wird ein neues Register
   * mit fortlaufender Nummerierung erzeugt.
   * @param regname Name des neuen Registers
   * @return tatschlich verwendeter Name
   */
  public String erzeugeRegister(String regname) {
    String generierterName = regname;
    boolean nameVorhanden;
    int count = 0;
    do {
      nameVorhanden = false;
      for(int i=0; i<registerListe.size(); i++)
        if(generierterName.equalsIgnoreCase(""+registerListe.get(i)))
          nameVorhanden = true;
      if(nameVorhanden)
        generierterName = regname + " (" + ++count + ")";
    } while(nameVorhanden);
    Register register = new Register(generierterName);
    // Register erzeugt, jetzt einsortieren:
    int size = registerListe.size();
    int pos = -1;
    for(int i=0; i<size; i++)
      if(generierterName.compareTo(""+registerListe.get(i)) >= 0)
        pos = i;
    if(pos == size-1)
      registerListe.add(register);
    else
      registerListe.add(pos+1, register);
    if(DEBUG)
      System.out.println("Register "+generierterName+" erzeugt.");
    geaendert = true;
    return generierterName;
  }

  /**
   * Liefert das passende Register zum angegebenen Namen.
   * @param regname Name des gesuchen Registers
   * @return gesuchtes Register
   */
  private Register findeRegister(String regname) {
    for(int i=0; i<registerListe.size(); i++) {
      if(regname.equals(""+registerListe.get(i)))
        return (Register)registerListe.get(i);
    }
    return null;
  }

  /**
   * Liefert das passende Register zum angegebenen Namen.
   * Wenn das Register noch nicht existiert, wird es erzeugt.
   * @param regname Name des gesuchen Registers
   * @return gesuchtes oder erzeugtes Register 
   */
	public Register findeOderErzeugeRegister(String regname) {
    Register register = findeRegister(regname);
    if(register == null) {
      erzeugeRegister(regname);
      register = findeRegister(regname);
    }
    return register;
  }
  
	/**
   * Liefert eine Liste mit alle Register-Namen.
   * Diese wird fr die Combo-Boxen zur Auswahl eines Registers bentigt.
	 * @return Array mit den Register-Namen
	 */
  public String[] getRegisterNamen() {
		int anzahl = registerListe.size();
		String[] namen = new String[anzahl];
		for(int i=0; i<anzahl; i++)
			namen[i] = ""+registerListe.get(i);
		return namen;
	}

  /**
   * bertrgt die Buchungen aus dem Quell-Register ins Ziel-Register.
   * Das Quell-Register wird anschlieend gelscht.
   * @param quelle Name des Quell-Registers
   * @param ziel Name des Ziel-Registers
   */
  public void registerVereinigen(String quelle, String ziel) {
    Register quellRegister = findeRegister(quelle);
    Register zielRegister = findeRegister(ziel);
    zielRegister.registerVereinigen(quellRegister);
    registerListe.remove(quellRegister);
    geaendert = true;
  }

	// -- Buchungen ---------------------------------------------

  /**
   * Erzeugt im angegebenen Register eine neue Buchung.
   * Dies ist in der Regel die noch leere Buchung am Ende des Registers.
   * @param regname Name des Registers
   */
  public int standardBuchungEinfuegen(String regname, StandardBuchung buchung) {
    Register register = findeRegister(regname);
    geaendert = true;
		cacheAktuell = false;
    return register.einsortierenBuchung(buchung);
  }

	/**
   * Erzeugt eine Umbuchung und fgt diese im Quell- und Zielregister ein.
	 * @param datum Buchungdatum
	 * @param buchungstext Buchungstext
	 * @param quelle Quellregister
	 * @param ziel Zielregister
	 * @param betrag Betrag
	 */
  public void umbuchungEinfuegen(Datum datum, String buchungstext, String quelle, String ziel, Euro betrag) {
    Register[] register = new Register[2];
    register[0] = findeRegister(quelle);
    register[1] = findeRegister(ziel);
    new Umbuchung(datum, buchungstext, register, betrag);
    // beim Erzeugen der Umbuchung wird diese automatisch in die beiden Register
    // eingefgt.
    geaendert = true;
  }
  
  /**
   * Liefert die Anzahl der Buchungen im angegebenen Register.
   * @param regname Name des Registers
   * @return Anzahl der Buchungen
   */
  public int getAnzahlBuchungen(String regname) {
    Register register = findeRegister(regname);
    if(register == null) {
      if(DEBUG)
        System.out.println("Datenbasis.getAnzahlBuchungen: "+regname+" gibt es nicht.");
      return 0;
    }
    return register.getAnzahlBuchungen();
  }

  /**
   * Entfernt eine Buchung aus einem Register.
   * @param regname Name des Registers
   * @param index Position der Buchung
   */
  public void entferneBuchung(String regname, int index) {
    Register register = findeRegister(regname);
    register.entferneBuchung(index);
    geaendert = true;
		cacheAktuell = false;
  }
  
  /**
   * Ersetzt eine Buchung durch eine andere.
   * @param regname Name des Registers
   * @param index Position der Buchung
   * @param buchung Neue Buchung
   */
  public void ersetzeBuchung(String regname, int index, AbstractBuchung buchung) {
		Register register = findeRegister(regname);
		register.entferneBuchung(index);
		register.einsortierenBuchung(buchung);
		geaendert = true;
		cacheAktuell = false;
  }

  /**
   * Liefert eine Buchung aus einem Register.
   * @param regname Name des Registers
   * @param index Position der Buchung
   */
  public AbstractBuchung getBuchung(String regname, int index) {
    Register register = findeRegister(regname);
    return register.getBuchung(index);
  }

  /**
   * Sortiert die angegebene Buchung neu ins Register ein.
   * Dies wird notwendig, wenn das Datum der Buchung gendert wurde.
   * @param regname Name des Registers
   * @param index Index der genderten Buchung
   * @return Einfge-Position 
   */
  public int buchungNeusortieren(String regname, int index) {
    Register register = findeRegister(regname);
    return register.buchungNeusortieren(index);
  }

	/**
   * Lscht alle Buchungen vor dem angegebenen Datum.
	 * @param datum Datum
	 */
  public void entferneAlteBuchungen(Datum datum) {
		int anzahlRegister = registerListe.size();
		int[] pos = new int[anzahlRegister];
		Euro[] salden = new Euro[anzahlRegister];
		// Erstmal die Salden aller Register ermitteln ...
		for(int i=0; i<anzahlRegister; i++) {
			Register register = (Register)registerListe.get(i);
			pos[i] = -1; // -1 = keine Buchungen lschen
			for(int j=0; j<register.getAnzahlBuchungen();j++)
				if(register.getBuchung(j).compareTo(datum) <= 0)
					pos[i] = j;
			if(pos[i] > -1)
				salden[i] = register.getSaldo(pos[i]);
		}
		// ... und dann ggf. alte Buchungen lschen und Selbstbuchung einfgen.
		for(int i=0; i<anzahlRegister; i++) {
			if(pos[i] > -1) {
				Register[] registerFuerSelbstbuchung = new Register[2];
				Register register = (Register)registerListe.get(i);
        register.removeBisBuchung(pos[i]);
				registerFuerSelbstbuchung[0] = register;
				registerFuerSelbstbuchung[1] = register;
				new Umbuchung((Datum)datum.clone(), "Erffnungssaldo", registerFuerSelbstbuchung, salden[i]);
				// Umbuchung werden automatisch einsortiert
				if(DEBUG)
				  System.out.println("-I- Im Register "+register+" wurden "+pos[i]+" Buchungen gelscht!");
			}
		}
		geaendert = true;
		cacheAktuell = false;
	}

  // -- Suchen und Ersetzen -----------------------------------

  private int registerSuchIdx = 0;
  private int buchungSuchIdx = 0;

  /**
   * Beginnt die Suche wieder in der ersten Buchung des ersten Registers.
   *
   */
  public void resetSuchIdx() {
    registerSuchIdx = 0;
    buchungSuchIdx = 0;
  }

  /**
   * Sucht nach dem angegebenen Text in allen Buchungen.
   * @param text gesuchter Text
   * @return Buchung in der der Text gefunden wurde
   */
  public AbstractBuchung suchen(String text, boolean grossUndKlein) {
    for(int i=registerSuchIdx; i<registerListe.size(); i++) {
      Register register = (Register)registerListe.get(i);
      for(int j=buchungSuchIdx; j<register.getAnzahlBuchungen(); j++) {
        AbstractBuchung buchung = register.getBuchung(j);
        if(buchung.sucheText(text, grossUndKlein)) {
          buchungSuchIdx = j+1; // da geht es weiter
          return buchung;
        }
      }
      buchungSuchIdx = 0;
      registerSuchIdx++;
    }
    registerSuchIdx = 0;
    return null;
  }
  
  /**
   * Liefert den Namen des Registers in der der Text gefunden wurde.
   * @return Register-Name
   */
  public String registerNameGefundenerText() {
    return ""+registerListe.get(registerSuchIdx);
  }

  /**
   * Liefert den Index der Buchung in der der Text gefunden wurde.
   * @return Index der Buchung
   */
  public int buchNrGefundenerText() {
    return buchungSuchIdx-1; // der Suchindex ist schon eine Position weiter
  }

  // -- Auswertungen ------------------------------------------
  
  private boolean unterkategorienVerwenden = true;

  /**
   * Legt fest, ob die Unterkatgorien bei Auswertungen verwendet weren sollen.
   * @param b <code>true</code> wenn sie verwendet werden
   */
  public void setUnterkategorienVerwenden(boolean b) {
    if(unterkategorienVerwenden == b)
      return;
    unterkategorienVerwenden = b;
    cacheAktuell = false;
  }
  
  /**
   * Liefert, ob die Unterkatgorien bei Auswertungen verwendet weren sollen.
   * @return <code>true</code> wenn sie verwendet werden
   */
  public boolean getUnterkategorienVerwenden() {
    return unterkategorienVerwenden;
  }
  
  /**
   * Liefert alle Buchungen einer Kategorie in einem Zeitraum.
   * @param zeitraum Zeitraum
   * @param regname Name des Registers
   * @param kategorien Liste mit Kategorien
   * @return Liste der Buchungen
   */
  public AbstractBuchung[] getBuchungen(AbstractZeitraum zeitraum, String regname, Kategorie[] kategorien) {
    ArrayList liste = new ArrayList();
    if(regname == null) {
	    for(int i=0; i<registerListe.size(); i++) {
	      Register register = (Register)registerListe.get(i);
	      getBuchungen(liste, zeitraum, register, kategorien);
	    }
    }
    else {
      Register register = findeRegister(regname);
      getBuchungen(liste, zeitraum, register, kategorien);
    }
    return (AbstractBuchung[])liste.toArray(new AbstractBuchung[liste.size()]);
  }
  
  private void getBuchungen(ArrayList liste, AbstractZeitraum zeitraum, Register register, Kategorie[] kategorien) {
    for(int i=0; i<register.getAnzahlBuchungen(); i++) {
      AbstractBuchung buchung = register.getBuchung(i);
      if(buchung.getDatum().istImZeitraum(zeitraum)) {
        for(int j=0; j<kategorien.length; j++) {
				 Euro wert = buchung.getKategorieWert(kategorien[j], unterkategorienVerwenden);
				 if(!wert.equals(Euro.NULL_EURO)) {
				   int anzahl = liste.size();
				   int pos = -1;
				   for(int k=0; k<anzahl; k++)
				     if(buchung.compareTo(liste.get(k)) >= 0)
				       pos = k;
				   if(pos == anzahl-1)  // neue Buchung ans Ende
				     liste.add(buchung);
				   else               // neue Buchung einfuegen
				     liste.add(pos+1, buchung);
				 }
        }
      }
    }
  }

  /**
   * Summiert alle Kategorien mit positivem Saldo im angegebenen
   * Register und Zeitraum auf.
   * @param regname Name der Registers, <code>null</code> = alle Register
   * @param zeitraum Zeitraum
   * @return Einnahmen
   */
  public Euro getEinnahmen(AbstractZeitraum zeitraum, String regname) {
    Euro einnahmen = new Euro();
    erneuereKategorieCache(zeitraum, regname);
    for(int i=0; i<kategorieListe.size(); i++) {
      Euro wert = ((Kategorie)kategorieListe.get(i)).getSumme();
      if(wert.compareTo(Euro.NULL_EURO) > 0)
        einnahmen.sum(wert);
    }
    return einnahmen;
  }

  /**
   * Summiert alle Kategorien mit negativem Saldo im angegebenen
   * Register und Zeitraum auf.
   * @param regname Name der Registers, <code>null</code> = alle Register
   * @param zeitraum Zeitraum
   * @return Ausgaben
   */
  public Euro getAusgaben(AbstractZeitraum zeitraum, String regname) {
    Euro ausgaben = new Euro();
    erneuereKategorieCache(zeitraum, regname);
    for(int i=0; i<kategorieListe.size(); i++) {
      Euro wert = ((Kategorie)kategorieListe.get(i)).getSumme();
      if(wert.compareTo(Euro.NULL_EURO) < 0)
        ausgaben.sum(wert);
    }
    return Euro.NULL_EURO.sub(ausgaben);
  }

	/**
	 * Ermittelt den Saldo an einem bestimmten Datum ber alle Register.
	 * @param datum Datum des gesuchten Saldos
	 * @return gesuchter Saldo
	 */
	public Euro getSaldo(Datum datum) {
		Euro saldo = new Euro();
		for(int i=0; i<registerListe.size(); i++) {
			Register register = (Register)registerListe.get(i);
			saldo.sum(register.getSaldo(datum));
		}
		return saldo;
	}

  /**
   * Liefert den Saldo an einer Position im Register.
   * @param regname Name des Registers
   * @param index Position
   */
  public Euro getRegisterSaldo(String regname, int index) {
    Register register = findeRegister(regname);
    return register.getSaldo(index);
  }

	/**
   * Liefert den Saldo an einem bestimmten Datum in einem bestimmten Register.
	 * @param regname Name der Registers
	 * @param datum Datum des gesuchten Saldos
	 * @return gesuchter Saldo
	 */
	public Euro getRegisterSaldo(String regname, Datum datum) {
		Register register = findeRegister(regname);
		return register.getSaldo(datum);
	}

  /**
   * Ermittelt den Saldo ber eine Kategorie in einem bestimmten Register und
   * Zeitraum.
   * @param regname Name des Registers, <code>null</code> = alle Register
   * @param kategorie Name der Kategorie
   * @param zeitraum Zeitraum
   * @return gesuchter Saldo
   */
  public Euro getKategorieSaldo(Kategorie kategorie, AbstractZeitraum zeitraum, String regname) {
    erneuereKategorieCache(zeitraum, regname);
    if(DEBUG)
      System.out.println(""+kategorie+": Saldo @ "+zeitraum+" = "+kategorie.getSumme());
    return kategorie.getSumme();
  }   

  /**
   * Liefert die Salden fr alle Kategorien in einem Zeitraum.
   * @param zeitraum Zeitraum
   * @param regname Name der Registers, <code>null</code> = alle Register
   * @return Salden aller Kategorien
   */
  public Euro[] getKategorieSalden(AbstractZeitraum zeitraum, String regname) {
    erneuereKategorieCache(zeitraum, regname);
    if(unterkategorienVerwenden) {
	    final int anzahlKategorien = kategorieListe.size();
	    Euro[] summen = new Euro[anzahlKategorien];
	    for(int i=0; i<anzahlKategorien; i++)
	      summen[i] = ((Kategorie)kategorieListe.get(i)).getSumme();
	    return summen;
    }
    // Nur die Hauptkategorien:
    return getKategorieSalden(getKategorien(), zeitraum, regname);
  }

  /**
   * Liefert die Salden fr bestimmte Kategorien in einem Zeitraum.
   * @param kategorien Kategorien
   * @param zeitraum Zeitraum
   * @param regname Name der Registers, <code>null</code> = alle Register
   * @return Salden der Kategorien
   */
  public Euro[] getKategorieSalden(Kategorie[] kategorien, AbstractZeitraum zeitraum, String regname) {
    erneuereKategorieCache(zeitraum, regname);
    Euro[] summen = new Euro[kategorien.length];
    for(int i=0; i<kategorien.length; i++)
      summen[i] = kategorien[i].getSumme();
    return summen;
  }

  private AbstractZeitraum zeitraumImCache;
  private String registerImCache;
  private boolean cacheAktuell = false;
  
  private void erneuereKategorieCache(AbstractZeitraum zeitraum, String regname) {
    // Testen ob (1) Zeitraum und (2) Registername dem Cache entsprechen und ob
    // (3) der cache noch aktuell ist
    if(((zeitraum == zeitraumImCache) || ((zeitraum != null) && zeitraum.equals(zeitraumImCache))) && 
       ((regname == registerImCache) || ((regname != null) && regname.equals(registerImCache))) && 
        cacheAktuell)
      return;
    zeitraumImCache = zeitraum;
    registerImCache = regname;
    cacheAktuell = true;
    final int anzahlKategorien = kategorieListe.size();
    final int anzahlRegister = registerListe.size();
		for(int i=0; i<anzahlKategorien; i++) {
			Kategorie kategorie = (Kategorie)kategorieListe.get(i);
			kategorie.loescheSumme();
		}
		if(regname == null)
			for(int i=0; i<anzahlRegister; i++) {
				final Register reg = (Register)registerListe.get(i);
				final int anzahlBuchungen = reg.getAnzahlBuchungen();
				for(int j=0; j<anzahlBuchungen; j++)
				  reg.getBuchung(j).bildeKategorieSumme(zeitraum, getUnterkategorienVerwenden());
	    }
		else {
	  	final Register register = findeRegister(regname);
			final int anzahlBuchungen = register.getAnzahlBuchungen();
			for(int j=0; j<anzahlBuchungen; j++)
			  register.getBuchung(j).bildeKategorieSumme(zeitraum, getUnterkategorienVerwenden());
		}
  }
  
  // -- gemerkte Buchungen ------------------------------------

	private final ArrayList gemerkteBuchungen = new ArrayList();
	private final ArrayList gemerkteBuchungenText = new ArrayList();

	/**
	 * Merkt sich die Standard- und SplitBuchungen. Falls es den Buchungstext schon
	 * gibt wird die zuvor gemerkte Buchung berschrieben.
	 * @param buchung Buchung zum Merken
	 */
	public void buchungMerken(AbstractBuchung buchung) {
		if(buchung.getClass() == Umbuchung.class)
		  return;
		// normale Buchungen merken (keine Umbuchungen!)
		int pos = Collections.binarySearch(gemerkteBuchungenText, buchung.getText());
		if (pos < 0) { // so eine Buchung gibt es noch nicht: Einfuegen
			gemerkteBuchungen.add(-pos-1, buchung);
			gemerkteBuchungenText.add(-pos-1, buchung.getText());
		}
		else {         // Buchung gibt es schon: Ueberschreiben
			gemerkteBuchungen.set(pos, buchung);
			gemerkteBuchungenText.set(pos, buchung.getText());
		}
	}

  /**
   * Liefert eine gemerkte Buchung, die mit dem angegebenen String beginnt.
   * @param prefix Anfang des gesuchten Buchungstextes
   * @return gefundene Buchung oder <code>null</code>, wenn nichts gefunden
   */
	public AbstractBuchung findeGemerkteBuchung(String prefix) {
		for(int i=0; i<gemerkteBuchungenText.size(); i++)
			if(((String)gemerkteBuchungenText.get(i)).startsWith(prefix))
				return (AbstractBuchung)gemerkteBuchungen.get(i);
		return null;
	}

  // -- Auto-Buchung ------------------------------------------
  
  private final String[] autoBuchungIntervallNamen =
  {"Monat","Quartal","Halbjahr","Jahr"};
  private final ArrayList autoBuchungen = new ArrayList();
  private final ArrayList autoBuchungRegister = new ArrayList();
  private final ArrayList autoBuchungIntervalle = new ArrayList();

  /**
   * Liefert die Intervall-Namen zur Initialisierung der Combo-Box.
   * @return Intervall-Namen
   */
  public String[] getAutoBuchungIntervallNamen() {
    return autoBuchungIntervallNamen;
  }
  
  /**
   * Liefert die Anzahl der vorhandenen wiederkehrenden Buchungen.
   * @return Anzahl
   */
  public int getAnzahlAutoBuchungen() {
    return autoBuchungen.size();
  }

	/**
	 * Liefert eine wiederkehrende Buchung.
   * @param index Nummer der Buchung
	 * @return Buchung
	 */
  public AbstractBuchung getAutoBuchung(int index) {
    return (AbstractBuchung)autoBuchungen.get(index);
  }

	/**
	 * Liefert das Register zu einer wiederkehrenden Buchung.
	 * @param index Nummer der Buchung
	 * @return Name des Registers
	 */
  public String getAutoBuchungRegister(int index) {
    return ""+autoBuchungRegister.get(index);
  }

  /**
   * Setzt das Register zu einer wiederkehrenden Buchung.
   * @param index Nummer der Buchung
   * @param register Name des Registers
   */
  public void setAutoBuchungRegister(int index, String register) {
    autoBuchungRegister.set(index, register);
  }

	/**
	 * Liefert das Intervall (Monat, Quartal, Halbjahr, Jahr) 
	 * zu einer wiederkehrenden Buchung.
	 * @param index Nummer der Buchung
	 * @return Name des Intervall
	 */
	public String getAutoBuchungIntervall(int index) {
  	return ""+autoBuchungIntervalle.get(index);
	}

	/**
	 * Setzt das Intervall (Monat, Quartal, Halbjahr, Jahr) 
	 * zu einer wiederkehrenden Buchung.
	 * @param index Nummer der Buchung
	 * @param intervall Name des Intervalls
	 */
	public void setAutoBuchungIntervall(int index, String intervall) {
		autoBuchungIntervalle.set(index, intervall);
	}

	/**
	 * Erzeugt eine neue wiederkehrenden Buchung.
	 * Dies geschied, wenn in der letzten Zeile der Tabelle eine Eingabe erfolgt.
	 */
	public void einfuegenNeueAutoBuchung() {
  	autoBuchungen.add(new StandardBuchung());
  	autoBuchungRegister.add(registerListe.get(0));
		autoBuchungIntervalle.add(autoBuchungIntervallNamen[0]);
	}

	/**
	 * Lscht eine wiederkehrende Buchung.
	 * @param index Nummer der Buchung
	 */
	public void entferneAutoBuchung(int index) {
	  if(index < autoBuchungen.size()) {
	    autoBuchungen.remove(index);
			autoBuchungRegister.remove(index);
			autoBuchungIntervalle.remove(index);
	  }
	}

	/**
	 * Fgt die flligen wiederkehrenden Buchungen in das entsprechende Register ein.
	 * @return Anzahl der eingefgten Buchungen
	 */
	public int autoBuchungenAusfuehren() {
	  int zaehler = 0;
	  for(int i=0; i<autoBuchungen.size(); i++) {
	    AbstractBuchung buchung = (AbstractBuchung)autoBuchungen.get(i);
	    Register register = findeRegister(""+autoBuchungRegister.get(i));
	    while(buchung.getDatum().compareTo(new Datum()) <= 0) {
	      register.einsortierenBuchung((AbstractBuchung)buchung.clone());
	      if(DEBUG)
	        System.out.println("AutoBuchung ausgefhrt: "+buchung.getText());
	      String intervall = (String)autoBuchungIntervalle.get(i);
	      buchung = (StandardBuchung)buchung.clone(); // sonst wrde die abgelegte Buchung gendert
	      buchung.setDatum(getFolgeDatum(buchung.getDatum(), intervall));
	      autoBuchungen.set(i, buchung);
	      zaehler++;
	    }
	  }
		if(zaehler > 0) {
			 geaendert = true;
				cacheAktuell = false;
		}			 
	  return zaehler;
	}

	/**
	 * Berechnet das nchste Datum an dem die wiederkehrende Buchung ausgefhrt
	 * werden soll.
	 * @param datum altes Datum
	 * @param intervall Intervall in den die Buchung ausgefhrt wird
	 * @return Neues Datum
	 */
	private Datum getFolgeDatum(Datum datum, String intervall) {
	  int monat = datum.getMonat();
	  int jahr = datum.getJahr();
	  if(intervall.equals(autoBuchungIntervallNamen[0])) // Monat
	    monat += 1;
	  if(intervall.equals(autoBuchungIntervallNamen[1])) // Quartal
	    monat += 3;
	  if(intervall.equals(autoBuchungIntervallNamen[2])) // Halbjahr
	    monat += 6;
	  if(intervall.equals(autoBuchungIntervallNamen[3])) // Jahr
	    jahr += 1;
	  if (monat > 12) {
	    monat -= 12;
	    jahr ++;
	  }
	  return new Datum(datum.getTag(), monat, jahr);
	}

  // -- Import ----------------------------------------------------------------
  
	/**
	 * Importiert eine CSV-Datei (Comma Separated Value) in das angegebene Register.
	 * Die bergebene Tabelle muss (mindestens) 4 Spalten besitzen.
	 * @param regname Name des Registers
	 * @param tabelle Tabelle
	 */
	public void importBuchungen(String regname, String[][] tabelle) {
		Register register = findeRegister(regname);
	  for(int i=0; i<tabelle.length; i++) {
	    Datum datum = new Datum(tabelle[i][0]);
	    String text = tabelle[i][1];
	    Kategorie kategorie = findeOderErzeugeKategorie(tabelle[i][2]);
	    Euro wert = new Euro(tabelle[i][3]);
	    AbstractBuchung buchung = new StandardBuchung(datum, text, kategorie, wert);
	    register.einsortierenBuchung(buchung);
	    geaendert = true;
			cacheAktuell = false;
	  }
	}
	
	/**
	 * Importiert eine Quicken-Export-Datei (QIF) in das angegebene Register.
	 * @param in Eingabe-Stream
	 * @param regname Name des Registers
	 * @param dm <code>true</code> wenn die Whrung der Daten D-Mark ist.
	 */
	public void importQuickenRegister(InputStream in, String regname, boolean dm) throws IOException, QuickenImportException {
	  Register register = findeRegister(regname);
	  String typ = leseQIFZeile(in);
	  if(DEBUG)
	    System.out.println("-I- QuickenImport: "+typ);
	  Register[] list = new Register[2];
	  list[1] = register; 
	  int c;
	  while((c = in.read()) != -1) { // EOF erreicht
	    // Datum einlesen
	    if(c != 'D')
	      throw new QuickenImportException("-E- QuickenImport: Ups! D (Datum) erwartet.");
	    StringTokenizer st = new StringTokenizer(leseQIFZeile(in), ".");
	    String monat = st.nextToken();
	    if(monat.length() == 1)
	      monat = "0" + monat;
	    String tag = st.nextToken();
	    if(tag.length() == 1)
	      tag = "0" + tag;
	    String jahr = st.nextToken();
	    Datum datum = new Datum(tag+"."+monat+"."+jahr);
	    
	    // Betrag einlesen
	    if(in.read() != 'U')
	      throw new QuickenImportException("-E- QuickenImport: Ups! U (Betrag) erwartet.");
	    Euro einzelWert = new Euro(leseQIFZeile(in).replace('.',',')); // US -> deutsch
	    if(in.read() != 'T')
	      throw new QuickenImportException("-E- QuickenImport: Ups! T (Betrag) erwartet.");
	    if(!einzelWert.equals(new Euro(leseQIFZeile(in).replace('.',','))))
	      throw new QuickenImportException("-E- QuickenImport: Ups! U und T Betrag ungleich.");
	    if(dm)
	      einzelWert.umrechnenVonDM();
	    
	    // Abgeglichene Buchung "X" einlesen
	    c = in.read();
	    if(c == 'C') {
	      leseQIFZeile(in);
	      c = in.read();
	    }
	    
	    // Buchungstext einlesen
	    String text = "*";
	    if(c == 'P') {
	      text = leseQIFZeile(in);
	      c = in.read();
	    }
	    
	    // Kategorie einlesen
	    String einzelKategorie;
	    if(c == 'L') {
	      einzelKategorie = leseQIFZeile(in);
	      c = in.read();
	    }
	    else
	      einzelKategorie = ""+Kategorie.SONSTIGES;
	    
	    // Split-Buchungen einlesen
	    StandardBuchung standardBuchung = null;
	    SplitBuchung splitBuchung = null;
	    while(c == 'S') {
	      String kategorie = leseQIFZeile(in);
	      if(in.read() != '$')
	        throw new QuickenImportException("-E- QuickenImport: Ups! $ (Betrag-Splitbuchung) erwartet.");
	      Euro wert = new Euro(leseQIFZeile(in).replace('.',',')); // US -> deutsch
	      if(dm)
	        wert.umrechnenVonDM();
	      // Prfen, ob Umbuchung:
	      if((kategorie.length() > 2) &&
	          kategorie.startsWith("[") &&
	          kategorie.endsWith("]")) {
	        list[0] = findeOderErzeugeRegister(kategorie);
	        if(wert.compareTo(Euro.NULL_EURO) > 0) {
	          new Umbuchung(datum, text, list, wert);
	          // Nur positive Umbuchungen erzeugen; Umbuchungen werden automatisch eingefgt
	          if(DEBUG)
	            System.out.println("Umbuchung "+text+" erzeugt.");
	        }
	      }
	      else { // keine Umbuchung:
	        if(standardBuchung == null) 
	          standardBuchung = new StandardBuchung(datum, text, findeOderErzeugeKategorie(kategorie), wert);
	        else if(splitBuchung == null) {
	          splitBuchung = new SplitBuchung(standardBuchung);
	          splitBuchung.add(findeOderErzeugeKategorie(kategorie), wert);
	        }
	        else
	          splitBuchung.add(findeOderErzeugeKategorie(kategorie), wert);
	      }
	      c = in.read();
	    } /* Ende while Split-Buchung*/
	    
	    if(c != '^')
	      throw new QuickenImportException("-E- QuickenImport: Ups! ^ (Datensatzende) erwartet.");
	    leseQIFZeile(in); // CR+LF von '^' lesen
	    
	    if(splitBuchung != null) {
	      register.einsortierenBuchung(splitBuchung);
        if(DEBUG)
          System.out.println("SplitBuchung erzeugt.");
	    }
	    else if(standardBuchung != null) {
	      register.einsortierenBuchung(standardBuchung);
        if(DEBUG)
          System.out.println("StandardBuchung erzeugt.");
	    }
	    else if((einzelKategorie.length() > 2) &&    // Pruefen ob Umbuchung
	        einzelKategorie.startsWith("[") &&
	        einzelKategorie.endsWith("]")) {
	      list[0] = findeOderErzeugeRegister(einzelKategorie);
	      if(einzelWert.compareTo(Euro.NULL_EURO) > 0) {
	        new Umbuchung(datum, text, list, einzelWert);
	        // Nur positive Umbuchungen erzeugen; Umbuchungen werden automatisch eingefgt
          if(DEBUG)
            System.out.println("Umbuchung "+text+" erzeugt.");
	      }
	    }
	    else {
	      register.einsortierenBuchung(new StandardBuchung(datum, text, findeOderErzeugeKategorie(einzelKategorie), einzelWert));
        if(DEBUG)
          System.out.println("StandardBuchung "+text+" erzeugt.");
	    }
	  } /* Ende while Datei einlesen*/

		cacheAktuell = false;
	  geaendert = true;
	}

	public class QuickenImportException
  extends Exception {
    public QuickenImportException(String text) {
      super(text);
    }
  }
    
	/**
	 * Liest eine einzelne Zeile aus der QIF-Datei (Quicken).
	 * @param in Eingabe-Stream
	 * @return gelesende Zeile
	 * @throws IOException
	 * @throws QuickenImportException
	 */
	private String leseQIFZeile(InputStream in)
		throws IOException, QuickenImportException {
		String zeile = "";
		int c;
		while((c = in.read()) != 13) { // Bis zum Zeilenende (CR)
			if(c == -1) {
				throw new QuickenImportException("-E- QuickenImport: Ups! Unerwartetes EOF.");
			}
			if(c != ',') // Kommata werden gefiltert; Problem beim Float-Parsen
				zeile = zeile + (char)c;
		}
		if((c = in.read()) != 10) {
			throw new QuickenImportException("-E- QuickenImport: Ups! Nach CR kam kein LF. "+(char)c);
		}
		return zeile;
	}

	// -- E/A-Funktionen -------------------------------------------------------
 
 /**
  * Ldt die Buchungen in die Register. Ldt und fhrt die wiederkehrenden Buchungen.
  */
 public void laden(DataInputStream in, String versionInfo) throws IOException {
	 // Register laden
	 int anzahl = in.readInt();
	 for(int i=0; i<anzahl; i++) {
		 Register register = findeOderErzeugeRegister(in.readUTF());
     register.laden(in, this);
	 }
	geaendert = false;
 	cacheAktuell = false;
	 if(DEBUG)
  	 System.out.println(""+anzahl+" Register geladen.");

	// automatische Buchungen laden und ausfhren (in v1.0 noch unbekannt!)
	if(!versionInfo.equals("jHaushalt1.0")) {
		int size = in.readInt();
		autoBuchungen.ensureCapacity(size);
		for(int i=0; i<size; i++) {
			// Buchung laden:
			AbstractBuchung buchung = null;
			String typ = in.readUTF();
			if(typ.equals("StandardBuchung")) {
				Datum datum = new Datum();
				datum.laden(in);
				String text = in.readUTF();
				int anz = in.readInt();
				if(anz == 1) {
					Kategorie kategorie = findeOderErzeugeKategorie(in.readUTF());
					Euro betrag = new Euro();
					betrag.laden(in);
					buchung = new StandardBuchung(datum, text, kategorie, betrag);
				}
				else
					throw new IOException("AutoBuchung: Falscher Buchungstyp: SplitBuchung");
			}
			else if(typ.equals("StandardBuchung2")) {
				buchung = new StandardBuchung();
				buchung.laden(in, this);
			}
			else
				throw new IOException("AutoBuchung: Falscher Buchungstyp: "+typ);
			// Register laden:
			String register = in.readUTF();
			findeOderErzeugeRegister(register);
			// Intervall laden:
			String zeitraum = in.readUTF();

			autoBuchungen.add(buchung);
			autoBuchungRegister.add(register);
			autoBuchungIntervalle.add(zeitraum);
		}
	 }
  }

  /**
   * Speichert die Register mit den Buchungen und die wiederkehrenden Buchung.
   * @param out Ausgabe-Stream
   * @throws IOException
   */
  public void speichern(DataOutputStream out)
	  throws IOException {
	  out.writeUTF("jHaushalt"+version);
	  out.writeInt(registerListe.size());
	  for(int i=0; i<registerListe.size(); i++) {
      Register register = (Register)registerListe.get(i);
      register.speichern(out);
    }
		out.writeInt(autoBuchungen.size());
		for(int i=0; i<autoBuchungen.size(); i++) {
			AbstractBuchung buchung = (AbstractBuchung)autoBuchungen.get(i);
			buchung.speichern(out);
			out.writeUTF(""+autoBuchungRegister.get(i));
			out.writeUTF(""+autoBuchungIntervalle.get(i));
		}
	  out.flush();
	  geaendert = false;
  }

}
